# -*- coding: utf-8 -*-
"""
@author: amakdisi@usgs.gov
"""

import json
import os

from typing import Union

import RiskTargetedGMCalc

from .GroundMotionHazard import GroundMotionHazard


class BuildingCodeRTGMCalc(object):
    """!
    Class for calculating risk-targeted ground motion values from a set of IM hazard curves, for a given construction codes document.
    """

    # Load building code risk targets mapping
    home_dir = os.path.dirname(os.path.realpath(__file__))

    with open(os.path.join(home_dir, "BldgCodeTargets.json")) as f:
        code_dict = json.load(f)

    @staticmethod
    def example_rtgm_calc(bldg_code: str = "ASCE7") -> dict:
        """!
        Static method for calculating risk-targeted motions using an example ground motion hazard curve
        @param bldg_code:String value referencing the construction code document for which to obtain the
                risk-targeted motions, ('ASCE7' or 'AASHTO')

        @ return Dictionary containing all site, risk target, summary, and detailed iteration calculation information
                 for all intensity measures
        """

        GM = GroundMotionHazard.from_examples()
        return BuildingCodeRTGMCalc.calc_rtgm(GM, bldg_code)

    @staticmethod
    def calc_rtgm(haz_data: Union[str, GroundMotionHazard], bldg_code: str) -> dict:
        """!
        Static method for calculating risk-targeted motions for a user-defined ground motion hazard curve
        @param haz_data: Dictionary of hazard curve data (see GroundMotionHazard docstring), or a path to a correctly-formatted .json or .csv file
        @param bldg_code:String value referencing the construction code document for which to obtain the
                risk-targeted motions, ('ASCE7' or 'AASHTO')

        @ return Dictionary containing all site, risk target, summary, and detailed iteration calculation information
                 for all intensity measures
        """

        # Parse the haz_data type
        # Direct input of dictionary or GMhazard class
        if type(haz_data) in [dict, GroundMotionHazard]:
            GM = haz_data
        # csv file
        elif haz_data.split(".")[-1] == "csv":
            GM = GroundMotionHazard.from_csv(haz_data)
        # json file
        elif haz_data.split(".")[-1] == "json":
            GM = GroundMotionHazard.from_json(haz_data)

        # Assign targets as class variables
        try:
            riskTargets = BuildingCodeRTGMCalc.code_dict[bldg_code]
        except (KeyError):
            raise KeyError(
                "BldCode string must be one of %s" % BuildingCodeRTGMCalc.code_dict.keys()
            )

        DATA = RiskTargetedGMCalc.calc_rtgm(
            haz_data=GM,
            retPeriod_uh=riskTargets["retPeriod_uh"],
            uh_time_yr=riskTargets["uh_time_yr"],
            PrC_tgt=riskTargets["PrC_tgt"],
            beta=riskTargets["beta"],
            tgtRisk_time=riskTargets["tgtRisk_time"],
            risk_time_yr=riskTargets["risk_time_yr"],
        )
        DATA["riskTargets"]["BldgCode"] = bldg_code

        return BldgCodeRTGM(DATA)


class BldgCodeRTGM(RiskTargetedGMCalc.RTGMData):
    """!
    Dictionary sub-class of compiled risk-targeted calculation data, inherited from RiskTargetedGMCalc module. Includes overloaded instance method for plotting results.
    """

    def __init__(self, rtgm_data):
        super().__init__(rtgm_data)

    def plotFullRiskCalc(self, imt: str = None):
        """!
        Overloaded instance method for plotting the full risk calculation results
        @param imt: Intensity measure tag (e.g. "PGA", "SA0P1")

        @returns matplotlib figure and subplot axes
        """
        fig, axs = super().plotFullRiskCalc(imt)
        fig.suptitle(
            self["site"]["name"] + "  ({})".format(self["riskTargets"]["BldgCode"]),
            fontsize=20,
            fontweight="bold",
            ha="center",
            y=0.985,
        )
        return fig, axs
