import csv
import json
import os
import random

from typing import NewType, Union

import matplotlib as mpl
import numpy as np
import scipy.interpolate as interpolate


GroundMotionHazard = NewType("GroundMotionHazard", dict)


class GroundMotionHazard(dict):

    """!
    Class for compiling or reading hazard curve data into dictionary format; for use as input into risk-targeted ground motion calculation
    """

    def __init__(
        self,
        site_name: str = None,
        lon: float = None,
        lat: float = None,
        Vs30: float = None,
        haz_curves: Union[tuple, list] = [],
        haz_curves_dict: dict = None,
    ) -> dict:
        """!
        @param site_name: string value denoting the site ID (optional, may be left as None)
        @param lat: site latitude in decimal degrees (optional, may be left as None)
        @param lon: site longitude in decimal degrees (optional, may be left as None)
        @param Vs30: Site Vs30 in m/s (optional, may be left as None)
        @param haz_curves: tuple or list of tuples containing hazard curve information for each intensity measure with the form (imt, iml, afe)
                    imt: string denoting intensity measure (e.g. "PGA", "SA0P1")
                    IML: list or 1-d array of IM level values
                    AFE: list or 1-d array of corresponding annualized exceedance rates (must be the same length as IML)
        @param haz_curves_dict: optional specification of hazard curve data via dictionary, with following structure
            keys:
                 imt:   string value denoting the intensity measure (e.g. 'PGA', 'SA1P01'),
                'iml':  1-D numpy array of im values
                'afe':  1-D numpy array of annualized frequencies of exceedance (must be the same length as 'iml')

        @return: Class instance of hazard curve data
        """
        # Create dictionary
        haz_dict = {
            "site": {"name": site_name, "lat": lat, "lon": lon, "Vs30": Vs30},
        }
        # Create hazard curve dictionary from input tuple
        haz_dict["hazCurves"] = {}
        # If user enters a single tuple, convert into list for iteration
        if type(haz_curves) == tuple:
            haz_curves = [haz_curves]
        for im in haz_curves:
            # Create nested dictionary and add to master
            if len(im[1]) != len(im[2]):
                raise ValueError(
                    "Hazard curve values for {} must have the same length, but have length {} and {}.".format(
                        im[0], len(im[1]), len(im[2])
                    )
                )
            # Potentially add in another check to ensure hazard curves have a minimum length?
            else:
                haz_dict["hazCurves"][im[0]] = {"iml": im[1], "afe": im[2]}
        # Additional logic allowing for direct input of hazard dictionary
        if len(haz_curves) == 0:
            if haz_curves_dict:
                haz_dict["hazCurves"] = haz_curves_dict
        super().__init__(haz_dict)

    @classmethod
    def from_dict(cls, haz_dict: dict) -> GroundMotionHazard:
        """!
        Class method to instantiate ground motion hazard curve using a dictionary
        @param haz_dict: Dictionary of the following structure:
                    keys:
                        site_name: string value denoting the site ID (optional, may be left as None)
                        lat: site latitude in decimal degrees (optional, may be left as None)
                        lon: site longitude in decimal degrees (optional, may be left as None)
                        Vs30: Site Vs30 in m/s (optional, may be left as None)
                        hazCurves: dictionary of hazard curve data, containing one or more sets of the following
                            keys:
                                imt:   string value denoting the intensity measure (e.g. 'PGA', 'SA1P01'),
                                'iml':  1-D numpy array of im values
                                'afe':  1-D numpy array of annualized frequencies of exceedance (must be the same length as 'iml')

        """
        return cls(
            site_name=haz_dict["site"]["name"],
            lat=haz_dict["site"]["lat"],
            lon=haz_dict["site"]["lon"],
            Vs30=haz_dict["site"]["Vs30"],
            haz_curves_dict=haz_dict["hazCurves"],
        )

    @classmethod
    def from_csv(cls, csv_path: str) -> GroundMotionHazard:
        """!
        Class method to instantiate ground motion hazard curve from a csv file
        @param: csv_path: string describing the path to the csv file

        @return: Class instance of hazard curve data
        """
        csv_data = []
        with open(csv_path) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            for row in csv_reader:
                csv_data.append(row)

        # Extract Site Data (if available)
        name = csv_data[0][1]
        lat = float(csv_data[1][1])
        lon = float(csv_data[2][1])
        vs30 = float(csv_data[3][1])

        # Extract hazard curve values
        num_ims = int(len(csv_data[5:]) / 3)

        haz_curve_list = []
        for im in range(num_ims):
            im_rows = csv_data[5 + 3 * im : 5 + 3 * im + 3]
            IMT = im_rows[0][0]
            # Convert IML and AFE value strings to floats
            IMLs = []
            AFEs = []
            for gm in zip(im_rows[1][1:], im_rows[2][1:]):
                # Exception handling to avoid reading blank strings
                try:
                    IMLs.append(float(gm[0]))
                    AFEs.append(float(gm[1]))
                except (ValueError):
                    pass
            haz_curve_list.append((IMT, IMLs, AFEs))

        return cls(site_name=name, lat=lat, lon=lon, Vs30=vs30, haz_curves=haz_curve_list)

    @classmethod
    def from_json(cls, json_path: str) -> GroundMotionHazard:
        """!
        Class method to instantiate ground motion hazard curve from json
        @param: json_path: string describing the path to the json file, which should have the following structure:
            keys:
                site_name: string value denoting the site ID (optional, may be left as None)
                lat: site latitude in decimal degrees (optional, may be left as None)
                lon: site longitude in decimal degrees (optional, may be left as None)
                Vs30: Site Vs30 in m/s (optional, may be left as None)
                hazCurves: dictionary of hazard curve data, containing one or more sets of the following
                    keys:
                         imt:   string value denoting the intensity measure (e.g. 'PGA', 'SA1P01'),
                        'iml':  1-D numpy array of im values
                        'afe':  1-D numpy array of annualized frequencies of exceedance (must be the same length as 'iml')

        @return: Class instance of hazard curve data
        """
        with open(json_path) as f:
            haz_dict = json.load(f)

        return cls(
            site_name=haz_dict["site"]["name"],
            lat=haz_dict["site"]["lat"],
            lon=haz_dict["site"]["lon"],
            Vs30=haz_dict["site"]["Vs30"],
            haz_curves_dict=haz_dict["hazCurves"],
        )

    @classmethod
    def from_examples(cls) -> GroundMotionHazard:
        """!
        Class method to instantiate ground motion hazard curve from library example files
        @return: Class instance of hazard curve data
        """
        home_dir = os.path.dirname(os.path.realpath(__file__))
        im_lengths = ["all-IMs", "short-IMs"]
        exampleHazDir = os.path.join(home_dir, "example-haz-inputs", random.choice(im_lengths))
        # Get the example .json hazard curve file names
        exFiles = [file for file in os.listdir(exampleHazDir) if ".json" in file]
        # Pick one at random and load into a dictionary
        hazEx = random.choice(exFiles)
        with open(os.path.join(exampleHazDir, hazEx)) as f:
            haz_dict = json.load(f)

        return cls(
            site_name=haz_dict["site"]["name"],
            lat=haz_dict["site"]["lat"],
            lon=haz_dict["site"]["lon"],
            Vs30=haz_dict["site"]["Vs30"],
            haz_curves_dict=haz_dict["hazCurves"],
        )

    def add_im_curves(self, imt: str, IML: Union[np.array, list], AFE: Union[np.array, list]):
        """!
        Instance method that modifies base class with hazard curve data for additional IMs
        @param imt: string denoting intensity measure (e.g. "PGA", "SA0P1")
        @param IML: list or 1-d array of IM level values
        @param AFE: list or 1-d array of corresponding annualized exceedance rates (must be the same length as IM)
        """
        self["hazCurves"][imt] = {"iml": IML, "afe": AFE}

    @staticmethod
    def upsample_single(
        iml: Union[np.array, list],
        afe: Union[np.array, list],
        anchor_im: float = None,
        min_scale: float = 0.10,
        max_scale: float = 10.0,
    ) -> dict:
        """!
        Static helper method for log-linear interpolation/extrapolation ground motion hazard curve to higher-resolution arrays
        @param iml: list or 1-D numpy array of IM level values
        @param afe: list or 1-D numpy array of corresponding annualized exceedance rates (must be the same length as IM)
        @param anchor_im: (float) Target IM value, used in setting re-sampling bounds (default is median)
        @param minScale: (float) Lower-bound scaling factor, pegged to uniform hazard value (default is 0.10)
        @param maxScale: (float) Upper-bound scaling factor, pegged to uniform hazard value (default is 10.0)

        @return Dictionary of the upsampled hazard curve, with keys "imlUP" and "afeUP" consisting of numpy arrays
        """

        # Convert to numpy arrays
        if type(iml) == list:
            iml = np.asarray(iml)

        if type(afe) == list:
            afe = np.asarray(afe)
        # Default anchor IM will just be median value
        if not anchor_im:
            anchor_im = np.exp(np.mean(np.log(iml)))

        # Set min and max bounds and check for inclusion of max and min raw values
        minGM = min(anchor_im * min_scale, iml.min())
        maxGM = max(anchor_im * max_scale, iml.max())

        # Set interpolation points based on min, max and log-increment
        gmInt = np.exp(np.arange(np.log(minGM), np.log(maxGM), np.log(1.05)))
        if gmInt.max() <= maxGM:
            gmInt = np.append(gmInt, maxGM)

        # Interpolation using linear extrapolate (to re-create current RTGM)
        tck = interpolate.interp1d(
            np.log(iml),
            np.log(afe),
            kind="linear",
            fill_value="extrapolate",
        )
        afeInt = np.exp(tck(np.log(gmInt)))
        out_data = {"imlUP": gmInt, "afeUP": afeInt}
        return out_data

    def upsample_all(
        self, anchor_uh_tr: float = 2475, min_scale: float = 0.10, max_scale: float = 10.0
    ):
        """!
        Instance method that interpolates and extrapolates all IM hazard curves in the GroundMotionHazard class
        @param anchor_uh_tr: (float) Uniform hazard return period at which anchoring IM value is calculated, used in setting re-sampling bounds (default is 2475 years)
        @param minScale: (float) Lower-bound scaling factor, pegged to uniform hazard value (default is 0.10)
        @param maxScale: (float) Upper-bound scaling factor, pegged to uniform hazard value (default is 10.0)

        @return Dictionary containing site information and upsampled hazard curve data for all IMs
        """

        # Get uniform hazard values as anchoring IMs
        uhgm = self.get_uh_iml(anchor_uh_tr)

        up_data = {"site": self["site"], "upsampled_hazard": {}}
        for im in self["hazCurves"]:
            # Extract raw hazard curve dictionary and anchoring IM value
            im_dict = self["hazCurves"][im]
            anchor_im = uhgm["UHGM"][im]["uh_gms"]
            up_data["upsampled_hazard"][im] = self.upsample_single(
                im_dict["iml"],
                im_dict["afe"],
                anchor_im=anchor_im,
                min_scale=min_scale,
                max_scale=max_scale,
            )

        return up_data

    def get_uh_iml(self, ret_periods: Union[list, float] = 2475) -> dict:
        """!
        Instance method that returns a dictionary of interpolated uniform hazard IM values
        @param ret_periods:

        @return Dictionary containing site information and interpolated, uniform IM values for each specified return period
        """
        if type(ret_periods) != list:
            ret_periods = [ret_periods]
        uhgm_dict = {}
        for imt in self["hazCurves"]:
            hazCurve = self["hazCurves"][imt]
            # Do this for one hazard curve first
            f = interpolate.interp1d(np.log(hazCurve["afe"]), np.log(hazCurve["iml"]))
            UHGM = np.exp(f(np.log(1 / np.asarray(ret_periods))))
            # Create results dictionary
            if len(UHGM) == 1:
                uhgm_dict[imt] = {"ret_periods": ret_periods[0], "uh_gms": UHGM[0]}
            else:
                uhgm_dict[imt] = {"ret_periods": ret_periods, "uh_gms": list(UHGM)}
        return {"site": self["site"], "UHGM": uhgm_dict}

    def plot_curves(
        self,
        ax: mpl.axes,
        imts: Union[str, list] = None,
        upsample: bool = True,
        anchor_uh_tr: float = 2475,
    ):
        """!
        Instance method that plots the ground motion hazard curves in log-space
        @param ax: matplotlib axis object
        @param imts: string or list of strings of IM tags
        @param upsample: (bool) flag indicating whether to plot the upsampled curves along with the raw data (default True)
        @param anchor_uh_tr: (float) Uniform hazard return period at which anchoring IM value is calculated, used in setting re-sampling bounds (default is 2475 years)
        """
        colors = ["royalblue", "crimson", "darkcyan", "orange", "mediumorchid", "gray"]
        markers = ["o", "s", ">", "^", "d"]
        # If no imts are provides, select up to five at random
        if not imts:
            all_imts = list(self["hazCurves"].keys())
            if len(all_imts) > 5:
                imts = sorted(random.sample(all_imts, 5))
            else:
                imts = sorted(all_imts)

        # Get anchoring IMs
        if upsample:
            uhgm = self.get_uh_iml(anchor_uh_tr)
            up_data = self.upsample_all(anchor_uh_tr)["upsampled_hazard"]

        ax.axhline(1 / anchor_uh_tr, color="gray", ls="--")
        for (im, col, mark) in zip(imts, colors, markers):
            if upsample:
                # Draw lines illustrating UHGM
                ax.axvline(uhgm["UHGM"][im]["uh_gms"], color="gray", ls="--")
                # Plot interpolated curves
                ax.plot(
                    up_data[im]["imlUP"], up_data[im]["afeUP"], lw=2, color=col, label="_nolegend"
                )
            ax.set_xscale("log")
            ax.set_yscale("log")
            ax.set_xlabel("IM Level")
            ax.set_ylabel(r"Annualized Exceedance ($\mathrm{yr}^{-1}$)")

            # Plot raw hazard curve data
            ax.plot(
                self["hazCurves"][im]["iml"],
                self["hazCurves"][im]["afe"],
                ls="",
                color=col,
                marker=mark,
                mfc="white",
                label=im,
            )

        # Set Axis title
        ax.text(
            0.98,
            0.98,
            self["site"]["name"],
            fontweight="bold",
            fontsize=14,
            transform=ax.transAxes,
            ha="right",
            va="top",
        )
        ax.legend(loc="lower left")

        return ax
