# -*- coding: utf-8 -*-
"""
@author: amakdisi@usgs.gov
Class for calculating risk-targeted ground motion values, anchored to a target conditional collapse probability
"""

from typing import Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy
import scipy.interpolate as interpolate
import scipy.stats as st

from cycler import cycler


mpl.rcParams["axes.prop_cycle"] = cycler(
    "color", ["royalblue", "crimson", "darkcyan", "orange", "mediumorchid", "gray"]
)


def calc_rtgm(
    haz_data: dict,
    retPeriod_uh: Union[float, int] = 2475,
    uh_time_yr: Union[float, int] = 50,
    PrC_tgt: float = 0.10,
    beta: float = 0.60,
    tgtRisk_time: float = 0.01,
    risk_time_yr: Union[float, int] = 50,
    errTol: float = 0.01,
    maxIter: int = 20,
) -> dict:
    """!
    Static method that iteratively calculates the risk-targeted ground motion values for all IM hazard curves
    @param haz_data: Dictionary of hazard curve data (see GroundMotionHazard docstring)
        keys:
            site_name: string value denoting the site ID (optional, may be left as None)
            lat: site latitude in decimal degrees (optional, may be left as None)
            lon: site longitude in decimal degrees (optional, may be left as None)
            Vs30: Site Vs30 in m/s (optional, may be left as None)
            hazCurves: dictionary of hazard curve data, containing one or more sets of the following
                keys:
                    imt:   string value denoting the intensity measure (e.g. 'PGA', 'SA1P01'),
                        'iml':  1-D numpy array of im values
                        'afe':  1-D numpy array of annualized frequencies of exceedance (must be the same length as 'iml')
    @param retPeriod_uh: Initial target return period (e.g. 1033 for AASHTO, 2475 for ASCE7)
    @param uh_time_yr: Time horizon for uniform hazard exceedance rate - used to calculate percent exceedance over time horizon
        (e.g. 75 yrs for AASHTO, 50 years for ASCE7)
    @param PrC_tgt: Target conditional collapse fragility percentile (in decimal form, e.g. 0.05 for AASHTO, 0.10 for ASCE7)
    @param beta: Standard deviation of fragility curve (e.g. 0.60)
    @param tgtRisk_time: Target risk of collapse over the specified time horizon (e.g. 0.015 [1.5%] for AASHTO, 0.01 [1%] for ASCE7)
    @param risk_time_yr: Time horizon over which collapse risk is to be determined (e.g. 75 years for AASHTO, 50 years for ASCE7)
    @param errTol:  (optional) Error below which the calculated RTGM is considered close enough to the target (default is 0.01)
    @param maxIter: (optional) Maximum number of iterations before algorithm terminates (default is 20)

    @return Dictionary containing all site, risk target, summary, and detailed iteration calculation information
    for all intensity measures
    """

    # Initialize final results dictionary
    DATA = {}

    # Create dictionary of site data
    DATA["site"] = haz_data["site"]

    # Create dictionary of risk targets
    risk_tgt_annual = -np.log(1 - tgtRisk_time) / risk_time_yr
    DATA["riskTargets"] = {
        "UH_retPeriod": retPeriod_uh,
        "condCollapseProb": PrC_tgt,
        "beta": beta,
        "tgtCollapseRisk_T": tgtRisk_time,
        "timeHorizon_Yr": risk_time_yr,
        "riskTgt_Annual": risk_tgt_annual,
    }

    # Add key for summary and RTGM data
    DATA["summary"] = {}
    DATA["RTGM"] = {}

    # Extract upsampled data
    haz_up = haz_data.upsample_all(retPeriod_uh)["upsampled_hazard"]
    # Extract uniform hazard im levels for the given return period
    uhgms = haz_data.get_uh_iml(retPeriod_uh)["UHGM"]
    # Go through each IM and perform RTGM calculation
    IMTs = list(haz_data["hazCurves"].keys())
    for im in IMTs:
        # Extract hazard curve data for given imt
        haz_im = haz_up[im]
        (iml, afe) = (np.asarray(haz_im["imlUP"]), np.asarray(haz_im["afeUP"]))
        # Initial guess im level is the uniform hazard value
        anchor_im = uhgms[im]["uh_gms"]
        # Run risk calculator for single im
        iter_data = calc_rtgm_single(
            iml,
            afe,
            anchor_im,
            PrC_tgt=PrC_tgt,
            beta=beta,
            tgtRisk_time=tgtRisk_time,
            risk_time_yr=risk_time_yr,
            errTol=errTol,
            maxIter=maxIter,
        )
        # Compiling results into dictionary
        rtgmFinal = iter_data[-1]["rtgmGuess"]
        rtgm_data = {
            "uhgm": anchor_im,
            "rtgm": rtgmFinal,
            "riskCoeff": rtgmFinal / anchor_im,
            "upsampledHazCurve": {"imlUP": list(iml), "afeUP": list(afe)},
            "iterations": iter_data,
        }
        # Creating summary dictionary of final values
        summary = {
            "uhgm": rtgm_data["uhgm"],
            "rtgm": rtgm_data["rtgm"],
            "riskCoeff": rtgm_data["riskCoeff"],
            "colRisk_annual": rtgm_data["iterations"][-1]["colRisk_Annual"],
            "timePeriod": risk_time_yr,
            "colRisk_T": rtgm_data["iterations"][-1]["colRisk_Yr"],
        }

        # Adding data to master dictionary
        DATA["summary"][im] = summary
        DATA["RTGM"][im] = {
            "riskTargets": DATA["riskTargets"],
            "rawHaz": haz_data["hazCurves"][im],
            "rtgmCalc": rtgm_data,
        }
    return RTGMData(DATA)

    # Returns a single dictionary of results for a single hazard curve
    # iml and afe should be upsampled!!!


def calc_rtgm_single(
    iml: np.array,
    afe: np.array,
    anchor_im: float,
    PrC_tgt: float,
    beta: float,
    tgtRisk_time: float,
    risk_time_yr: Union[int, float],
    errTol: float = 0.01,
    maxIter: int = 20,
) -> dict:
    """!
    Static helper method that iteratively calculates the risk-targeted ground motion for a single IM hazard curve
    @param 'iml':  1-D numpy array of im level values (should be upsampled)
    @param 'afe':  1-D numpy array of annualized frequencies of exceedance (must be the same length as 'iml') (should be upsampled)
    @param anchor_im: im level at which the initial rtgm guess is made, and the initial collapse fragility curve is anchored (usually the uniform hazard level)
    @param PrC_tgt: Target conditional collapse fragility percentile (in decimal form, e.g. 0.05 for AASHTO, 0.10 for ASCE7)
    @param beta: Standard deviation of fragility curve (e.g. 0.60)
    @param tgtRisk_time: Target risk of collapse over the specified time horizon (e.g. 0.015 [1.5%] for AASHTO, 0.01 [1%] for ASCE7)
    @param risk_time_yr: Time horizon over which collapse risk is to be determined (e.g. 75 years for AASHTO, 50 years for ASCE7)
    @param errTol:  (optional) Error below which the calculated RTGM is considered close enough to the target (default is 0.01)
    @param maxIter: (optional) Maximum number of iterations before algorithm terminates (default is 20)

    @return Dictionary containing detailed results of the iterative calculation
    """

    # Create dictionary of risk targets
    risk_tgt_annual = -np.log(1 - tgtRisk_time) / risk_time_yr

    # Create empty list to track results over all iterations
    rtgm_data = []
    colCDF = None
    for nIter in range(maxIter):
        # First iteration - UHS value is set as first guess
        if nIter == 0:
            IMval = anchor_im
        # Second iteration, IM guess is based on ratio between target and 1st-guess collapse risks
        elif nIter == 1:
            # Checking against target
            err = colCDF[-1] / risk_tgt_annual
            if np.abs(err - 1) <= errTol:
                break
            # Next guess
            IMval = err * IMval

        # 3+ iteration - Log-linearly interpolate afterwards
        else:
            # Checking against target
            err = colCDF[-1] / risk_tgt_annual
            if np.abs(err - 1) <= errTol:
                break
            # Otherwise, we interpolate
            # extracting annualized collapse risk and rtgm guesses
            colRisks = np.asarray([it["colRisk_Annual"] for it in rtgm_data])
            IMguesses = np.asarray([it["rtgmGuess"] for it in rtgm_data])
            xp = np.log(colRisks)
            yp = np.log(IMguesses)
            f = interpolate.interp1d(xp, yp)
            IMval = np.exp(f(np.log(risk_tgt_annual)))

        # Calculate the collapse risk for a single iteration
        (IMval, frCDF, frPDF, hazXpdf, colCDF, colRisk_Yr,) = calc_collapse_risk(
            iml, afe, anchor_im=IMval, PrC_tgt=PrC_tgt, beta=beta, risk_time_yr=risk_time_yr
        )

        colCDF_Yr = 1 - (1 - colCDF) ** risk_time_yr
        calcRiskAnnual = colCDF[-1]

        # Create results dictionary for each iteration
        dictLabels = [
            "iter",
            "rtgmGuess",
            "frCDF",
            "frPDF",
            "hazXpdf",
            "colCDF",
            "colCDF_Yr",
            "colRisk_Annual",
            "colRisk_Yr",
        ]
        dictVars = [
            nIter + 1,
            IMval,
            frCDF,
            frPDF,
            hazXpdf,
            colCDF,
            colCDF_Yr,
            calcRiskAnnual,
            colRisk_Yr,
        ]
        iterDict = {}
        for data in zip(dictLabels, dictVars):
            if type(data[1]) is np.ndarray:
                iterDict[data[0]] = list(data[1])
            else:
                iterDict[data[0]] = data[1]
        rtgm_data.append(iterDict)

    return rtgm_data


def calc_collapse_risk(
    iml: np.array,
    afe: np.array,
    anchor_im: float,
    PrC_tgt: float,
    beta: float,
    risk_time_yr: Union[float, int],
) -> tuple:
    """!
    Static helper method that performs a single risk calculation, based on generating a collapse fragility curve anchored to a user-defined intensity measure and target collapse probability
    @param 'iml':  1-D numpy array of im level values (should be upsampled)
    @param 'afe':  1-D numpy array of annualized frequencies of exceedance (must be the same length as 'iml') (should be upsampled)
    @param anchor_im: The intensity measure level at which the fragility curve is anchored - default is None, which uses the uniform hazard ground motion level
    @param PrC_tgt: Target conditional collapse fragility percentile (in decimal form, e.g. 0.05 for AASHTO, 0.10 for ASCE7)
    @param beta: Standard deviation of fragility curve (e.g. 0.60)
    @param risk_time_yr: Used to calculate the risk over a specified time horizon (e.g. 75 years for AASHTO, 0.10 for ASCE7).

    @return (anchorIM,frCDF,frPDF,hazXpdf,colCDF,colRisk_Y):
        anchorIM: (float) ground motion value used to anchor the fragility curve
        frCDF: (numpy array) the fragility curve (also can be described as the cumulative distribution function of the collapse capacity)
        frPDF: (numpy array) the probability density function of the collapse capacity
        hazXpdf: (numpy array) the product of the upsampled hazard curve and collapse capacity PDF
        colCDF: (numpy array) the cumulative collapse risk (over the specified time horizon)
        colRisk_Y: (float) the final collapse risk (over the specified time horizon)
    """
    # Calculating collapse fragility
    (iml, frPDF, frCDF) = calc_fragility(iml, anchor_im, PrC_tgt, beta)

    # Calculating collapse risk
    (iml, hazXpdf, colCDF, colRisk_Yr) = integrate_collapse_risk(iml, afe, frPDF, risk_time_yr)

    return (anchor_im, frCDF, frPDF, hazXpdf, colCDF, colRisk_Yr)


def calc_fragility(iml: np.array, anchor_im: np.array, PrC_tgt: float, beta: float) -> tuple:
    """!
    Static helper method for generating collapse capacity distribution, given a target GM level, anchoring percent collapse, and lognormal standard deviation
    @param 'iml':  1-D numpy array of im level values
    @param anchorIM: Target IM value at which the collapse fragility is anchored (defaults to self.UHGM if None is specified)
    @param PrC_tgt: Target conditional collapse fragility percentile (in decimal form, e.g. 0.05 for AASHTO, 0.10 for ASCE7)
    @param beta: Standard deviation of fragility curve (e.g. 0.60)

    @returns (im, pdf, cdf)
        im: (numpy array) upsampled ground motion amplitudes
        pdf: (numpy array) the probability density function of the collapse capacity
        cdf: (numpy array) the fragility curve (also can be described as the cumulative distribution function of the collapse capacity)
    """
    # Calculate log-median from percentile value
    logMed = np.log(anchor_im) - st.norm.ppf(PrC_tgt) * beta

    # Calculate pdf and cdf
    pdf = st.norm.pdf(np.log(iml), logMed, beta)
    cdf = st.norm.cdf(np.log(iml), logMed, beta)

    return (iml, pdf, cdf)


def integrate_collapse_risk(
    iml: np.array, afe: np.array, fragPDF: np.array, risk_time_yr: Union[float, int]
) -> tuple:
    """!
    Static helper method for calculating the unconditional collapse risk via numerical integration of the ground motion hazard curve and the user-defined collapse capacity
    @param 'iml':  1-D numpy array of im level values (should be upsampled)
    @param 'afe':  1-D numpy array of annualized frequencies of exceedance (must be the same length as 'iml') (should be upsampled)
    @param fragPDF: 1-D numpy array of collapse capacity densities, must have the same number of elements as im
    @param risk_time_yr: Time horizon over which collapse risk is to be calculated:
        For NEHRP/ASCE7: 50 years
        For AASHTO: 75 years

    @return (im,hazXpdf,colCDF,colRisk_Y):
        im: (numpy array) upsampled ground motion amplitudes
        hazXpdf: (numpy array) the product of the upsampled hazard curve and collapse capacity PDF
        colCDF: (numpy array) the cumulative collapse risk (over the specified time horizon)
        colRisk_Y: (float) the final collapse risk (over the specified time horizon)
    """
    # Calculating annual collapse frequency density (product of hazard curve and collapse PDF)
    hazXpdf = fragPDF * afe

    # Numerical integration
    colCDF = scipy.integrate.cumtrapz(hazXpdf, np.log(iml), initial=0)
    colProb = scipy.integrate.trapz(hazXpdf, np.log(iml))

    # Calculate collapse risk over specified time horizon
    #colRisk_Yr = 1 - (1 - colProb) ** risk_time_yr
    colRisk_Yr = 1 - np.exp(- colProb * risk_time_yr)
    return (iml, hazXpdf, colCDF, colRisk_Yr)


class RTGMData(dict):
    """!
    Dictionary sub-class of compiled risk-targeted calculation data. Includes instance method for plotting results. Used by RiskTargetedGMCalc.rtgm_calc method.
    """

    def __init__(self, rtgm_data):
        super().__init__(rtgm_data)

    def plotFullRiskCalc(self, imt: str = None) -> tuple:
        """!
        Instance method for plotting the full risk calculation results
        @param imt: Intensity measure tag (e.g. "PGA", "SA0P1")

        @returns matplotlib figure and subplot axes
        """
        if not imt:
            imt = list(self["RTGM"].keys())[0]

        data = self["RTGM"][imt].copy()
        fig, axs = plt.subplots(5, 1)
        fig.set_size_inches(10, 15)
        fig.set_tight_layout("tight")
        fig.suptitle(
            self["site"]["name"],
            fontsize=20,
            fontweight="bold",
            ha="center",
            y=0.985,
        )
        ylabels = [
            "Annual Frequency\nof Exceedance",
            "Conditional Collapse Probability",
            "Conditional Collapse\nProbability Density",
            "Annual Collapse\nFrequency Density",
            "Cumulative %s-Year\nCollapse Probability" % data["riskTargets"]["timeHorizon_Yr"],
        ]

        varTags = ["hazCurve", "frCDF", "frPDF", "hazXpdf", "colCDF_Yr"]
        titles = [
            "Hazard Curve",
            "Fragility Curves",
            "Derivative of Fragility Curves",
            "Hazard Curve $\\times$\nDerivative of Fragility Curves",
            "Cumulative Integral of Hazard Curve $\\times$\n Derivative of Fragility Curves",
        ]

        if imt[0] == "P":
            if imt[-1] == "A":
                xlabel = "Peak Ground Acceleration (g)"
            else:
                xlabel = "Peak Ground Velocity (m/s)"
        else:
            saT = float(imt.split("SA")[-1].replace("P", "."))
            xlabel = r"Spectral Acceleration, $S_a({}\mathrm{{s}})$ (g)".format(saT)

        for a in enumerate(axs[0:5]):
            ax = axs[a[0]]
            ax.grid(True, alpha=0.25)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabels[a[0]])
            ax.set_xscale("log")
            if a[0] not in [1, 2, 3, 4]:
                ax.set_yscale("log")
            ax.set_title(titles[a[0]], x=0.01, y=0.05, ha="left", fontsize=14, fontweight="bold")
            # Plot IM hazard curve first
            if a[0] == 0:
                # Draw lines illustrating UHGM
                ax.axvline(data["rtgmCalc"]["uhgm"], color="gray", ls="--")
                ax.axhline(1 / data["riskTargets"]["UH_retPeriod"], color="gray", ls="--")

                # Plot hazard curves
                hazInt = data["rtgmCalc"]["upsampledHazCurve"]
                hazCurve = data["rawHaz"]
                ax.plot(
                    hazInt["imlUP"],
                    hazInt["afeUP"],
                    lw=2.5,
                    color="black",
                    label="Upsampled",
                )
                ax.plot(
                    hazCurve["iml"],
                    hazCurve["afe"],
                    ls="",
                    marker="o",
                    mfc="white",
                    markersize=8,
                    mew=2,
                    color="black",
                    label="Raw Data",
                )
                ax.legend(loc="upper right", fontsize=12)
                # Text for exceedance rate
                pctUH = 1 - np.exp(
                    -data["riskTargets"]["timeHorizon_Yr"] / data["riskTargets"]["UH_retPeriod"]
                )
                (yMin, yMax) = np.log(ax.get_ybound())
                textY = (np.log(1 / data["riskTargets"]["UH_retPeriod"]) - yMin) / (yMax - yMin)
                ax.text(
                    0.01,
                    textY - 0.02,
                    "%d%% Probability of Exceedance in %d Years"
                    % (np.round(100 * pctUH), data["riskTargets"]["timeHorizon_Yr"]),
                    fontsize=11,
                    transform=ax.transAxes,
                    ha="left",
                    va="top",
                )
            else:
                im = np.asarray(hazInt["imlUP"])
                # Loop through each iteration
                for it in data["rtgmCalc"]["iterations"]:
                    ax.plot(im, it[varTags[a[0]]], lw=2, label="Iteration %d" % it["iter"])
                ax.legend(loc="upper left", fontsize=12)
                if a[0] == 4:
                    ax.axhline(self["riskTargets"]["tgtCollapseRisk_T"], color="gray", ls="--")
                    # Text for target collapse risk
                    (yMin, yMax) = ax.get_ybound()
                    riskTgtYr = data["riskTargets"]["tgtCollapseRisk_T"]
                    timeYr = data["riskTargets"]["timeHorizon_Yr"]
                    textY = (riskTgtYr - yMin) / (yMax - yMin)
                    ax.text(
                        0.01,
                        textY - 0.02,
                        "%0.1f%% Probability of Collapse in %d Years" % (100 * riskTgtYr, timeYr),
                        fontsize=11,
                        transform=ax.transAxes,
                        ha="left",
                        va="top",
                    )
                    ax.legend(loc="lower right", fontsize=12)
        return fig, axs
